@@ -2,7 +2,7 @@
 DSLIBRARY - run your data science code anywhere.
 
 """
-__version__ = "0.0.4"
+__version__ = "0.0.6"
 
 # environment variables
 from .front import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TOKEN
@@ -10,6 +10,8 @@ from .front import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TOKEN
 from .front import DSLibrary
 # run models
 from .run_models import ModelRunner
+# utils for doing setup
+from .utils.instantiation import parse_verb_and_args, instantiate_class_and_args
 
 # if a package called 'mmlibrary' is installed it becomes the default target
 try:
@@ -48,27 +50,22 @@ def instance(env: dict=None):
     from .transport.to_volume import DSLibraryViaVolume
     from .transport.to_mmlibrary import DSLibraryViaMMLibrary
 
-    import importlib
     import json
     import os
     if env is None:
         env = os.environ
-    spec = json.loads(env.get(ENV_DSLIBRARY_SPEC) or "{}")
+    spec = env.get(ENV_DSLIBRARY_SPEC) or "{}"
+    if isinstance(spec, str):
+        spec = json.loads(spec)
     target = env.get(ENV_DSLIBRARY_TARGET) or ("mmlibrary" if mmlibrary else "local")
     target_parts = target.split(":")
     # point to a REST API
     if target_parts[0] in ("http", "https"):
         token = env.get(ENV_DSLIBRARY_TOKEN)
         return DSLibraryViaREST(url=target, token=token)
+    # FIXME protocol handlers need to use 'env'
     # gather arguments
-    args = []
-    kwargs = {}
-    for arg in target_parts[1:]:
-        arg_parts = arg.split('=', maxsplit=1)
-        if len(arg_parts) == 1:
-            args.append(arg)
-        else:
-            kwargs[arg_parts[0]] = arg_parts[1]
+    mode, args, kwargs = parse_verb_and_args(target)
     if spec:
         kwargs["spec"] = spec
     # a strictly local instance
@@ -81,13 +78,7 @@ def instance(env: dict=None):
     if target_parts[0] == "volume":
         return DSLibraryViaVolume(*args, **kwargs)
     # we fall through to custom implementation support
-    cls_parts = target_parts[0].split(".")
-    try:
-        module = importlib.import_module('.'.join(cls_parts[:-1]))
-        cls = getattr(module, cls_parts[-1])
-        return cls(*args, **kwargs)
-    except ImportError:
-        raise ValueError(f"Unrecognized {ENV_DSLIBRARY_TARGET} implementation: {target_parts[0]}")
+    return instantiate_class_and_args(target, DSLibrary)
 
 
 # set up default methods
