@@ -15,9 +15,10 @@ import typing
 from collections import namedtuple
 import yaml
 
+from .proto_handlers import load_protocol_handler
 from .utils.connect import connect_to_database
 from .utils.dbconn import sql_enquote_id
-from .utils.file_utils import write_stream_with_read_on_close
+from .utils.file_utils import write_stream_with_read_on_close, connect_to_filesystem
 
 try:
     import mlflow
@@ -56,7 +57,9 @@ class DSLibrary(object):
                         methods to MLFlow.
         """
         self._spec = spec or {}
-        jsonschema.validate(self._spec, LOCAL_SPECS_SCHEMA)
+        spec_wo_data = dict(self._spec)
+        spec_wo_data.pop("data", None)
+        jsonschema.validate(spec_wo_data, LOCAL_SPECS_SCHEMA)
         self._run_id = self._spec.get("run_id") or ""
         self._mlflow_opts = self._spec.get("mlflow") or {}
         self._mlflow_all = self._mlflow_opts.get("all")
@@ -65,6 +68,18 @@ class DSLibrary(object):
             raise DSLibraryException("MLFlow was selected for output but the mlflow package is not installed")
         self._mlflow_run = None
 
+    def _xlt_resource(self, resource_name, _mode="inputs", **kwargs):
+        """
+        Look up specifications for a named input or output and use those to override supplied defaults.
+        :param resource_name:   Named input.
+        :param kwargs:          Defaults.
+        :return:    A tuple with (0) the URI, and (1) the final set of options
+        """
+        resource_spec = (self._spec.get(_mode) or {}).get(resource_name) or {}
+        kwargs.update(resource_spec)
+        uri = kwargs.pop("uri", None) or resource_name
+        return uri, kwargs
+
     def get_metadata(self) -> Metadata:
         """
         Load self-descriptive metadata for this model.  Contains information about parameters, inputs, & outputs.
@@ -164,6 +179,8 @@ class DSLibrary(object):
         given a name.  The caller selects particular data sources and associates them with these names.  The burden of
         indicating file format details is on the caller.
 
+        A good practice is to specify a logical name in 'resource_name', and provide a default uri in 'uri'.
+
         NOTES:
           * Inputs and outputs may be segregated: writing to an output named 'x' may have no effect on the input
             named 'x'.
@@ -178,35 +195,24 @@ class DSLibrary(object):
 
         :returns:       A file-like object.
         """
-        open_args = dict(kwargs)
-        if "://" in resource_name:
-            # an external (untracked) resource
-            path = resource_name
-        elif resource_name.startswith("./"):
-            # a local file - the "./" bypasses lookup of named inputs
-            path = resource_name
-        else:
-            # a named resource
-            resource_spec = (self._spec.get("inputs" if 'r' in mode else "outputs") or {}).get(resource_name)
-            if resource_spec:
-                # options come from mapping, and override supplied options
-                open_args.update(resource_spec)
-                path = open_args.pop("uri", resource_name)
-            else:
-                # if there is no mapping we look for a local file
-                path = resource_name
+        uri, open_args = self._xlt_resource(resource_name, _mode="inputs" if "r" in mode else "outputs", **kwargs)
+        # remove formatting arguments
+        open_args.pop("format", None)
+        open_args.pop("format_options", None)
         # send all outputs to mlflow if requested
         use_mlflow = self._mlflow_all and ("w" in mode or "a" in mode)
         if use_mlflow:
             if "a" in mode:
                 raise DSLibraryException("Append mode not available with mlflow output")
-            return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=lambda fh: mlflow.log_artifact(fh.name, resource_name))
-        # remove formatting arguments
-        open_args.pop("format_options", None)
+            return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=lambda fh: mlflow.log_artifact(fh.name, uri))
+        # check for a custom protocol handler
+        proto_handler = load_protocol_handler(uri)
+        if proto_handler:
+            return proto_handler.open_resource(uri, mode, **open_args)
         # 'path' has the URI or path of what we need to open
         # 'mode' has the open mode
         # 'open_args' has additional arguments to pass
-        return self._opener(path, mode, **open_args)
+        return self._opener(uri, mode, **open_args)
 
     def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
         """
@@ -235,6 +241,40 @@ class DSLibrary(object):
         """
         self.log_dict({"uri": self.get_uri(), "success": success, **kwargs}, EVALUATION_RESULT_ALIAS)
 
+    def _get_system_connection(self, system_type, resource_name: str, for_write: bool=False, **kwargs):
+        """
+        Access all types of external system connections.
+        :param system_type:         See methods below.
+        :param resource_name:       Named input or output, or URI.
+        :param for_write:           Request write access.
+        :param kwargs:              Custom arguments.
+        :return:        An engine implementation, the type of which depends on the requested system_type.
+                        See 'engine_intf'.
+        """
+        # translate 'resource_name' based on supplied specification and overriding supplied arguments
+        uri, kwargs = self._xlt_resource(resource_name, "outputs" if for_write else "inputs", **kwargs)
+        handler = load_protocol_handler(uri)
+        if handler:
+            return handler.get_system_connection(system_type, uri=resource_name, for_write=for_write, **kwargs)
+        if system_type == "filesystem":
+            return connect_to_filesystem(uri=uri, for_write=for_write, **kwargs)
+        if system_type == "sql":
+            library = kwargs.pop("library", None)
+            return connect_to_database(uri, library=library, for_write=for_write, **kwargs)
+        # TODO implement nosql connection
+        # TODO implement timeseries connection
+        raise ValueError(f"Unsupported system_type: {system_type}")
+
+    def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
+        """
+        Returns an 'open(filename, mode)' method for a given external filesystem, like s3, abs, gcs, etc..
+
+        :param resource_name:   Which filesystem to access.  This can be the name of an input or output resource, or a
+                                URI providing connection details.
+        :param for_write:       True to enable write operations.
+        """
+        return self._get_system_connection("filesystem", resource_name, for_write, **kwargs)
+
     def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
         """
         Returns a DBI-compatible connection object which can be used to communicate with a relational database.
@@ -243,14 +283,9 @@ class DSLibrary(object):
                                 URI referencing an external data source.
         :param for_write:       True to enable write operations.
         """
-        # translate 'resource_name' based on supplied specification and overriding supplied arguments
-        resource_spec = (self._spec.get("outputs" if for_write else "inputs") or {}).get(resource_name)
-        kwargs.update(resource_spec)
-        # connect to data
-        uri = kwargs.pop("uri", None)
-        return connect_to_database(uri=uri, for_write=for_write, **kwargs)
+        return self._get_system_connection("sql", resource_name, for_write, **kwargs)
 
-    def get_nosql_connection(self, resource_name: str, **kwargs):
+    def get_nosql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
         """
         Returns a connection to a NoSQL database, as an interface that abstracts the functionality of a range of
         popular NoSQL engines.
@@ -258,8 +293,7 @@ class DSLibrary(object):
         :param resource_name:  Which database to access.
             This can be the name of an input or output resource, or a URI referencing an external data source.
         """
-        # TODO design and implement this
-        raise DSLibraryException("not yet implemented")
+        return self._get_system_connection("nosql", resource_name, for_write, **kwargs)
 
     def get_timeseries_db_connection(self, resource_name: str, for_write: bool=False, **kwargs):
         """
@@ -271,8 +305,7 @@ class DSLibrary(object):
         :return:    A connection, as a yet-to-be-determined interface that supports a range of popular time series
                     databases.
         """
-        # TODO design and implement this
-        raise DSLibraryException("not yet implemented")
+        return self._get_system_connection("timeseries", resource_name, for_write, **kwargs)
 
     def log_param(self, key: str, value):
         """
@@ -308,17 +341,26 @@ class DSLibrary(object):
             if mlflow:
                 mlflow.log_metric(metric_name, metric_value, step)
             return
-        row = {
-            "uri": self.get_uri(),
-            "run_id": self._spec.get("run_id") or "",
-            "user": self._spec.get("user") or "",
-            "time": time.time(),
-            "name": metric_name,
-            "value": metric_value,
-            "step": step
-        }
-        df = pandas.DataFrame(data=[row])
-        self.write_resource(METRICS_ALIAS, df, append=True)
+        if self._capturing_output(METRICS_ALIAS):
+            # captured metrics only return a name and value
+            if "capture" not in self._spec:
+                self._spec["capture"] = {}
+            if METRICS_ALIAS not in self._spec["capture"]:
+                self._spec["capture"][METRICS_ALIAS] = {}
+            self._spec["capture"][METRICS_ALIAS][metric_name] = metric_value
+        else:
+            # full metrics output
+            row = {
+                "uri": self.get_uri(),
+                "run_id": self._spec.get("run_id") or "",
+                "user": self._spec.get("user") or "",
+                "time": time.time(),
+                "name": metric_name,
+                "value": metric_value,
+                "step": step
+            }
+            df = pandas.DataFrame(data=[row])
+            self.write_resource(METRICS_ALIAS, df, append=True)
 
     def log_metrics(self, metrics: dict, step: int=0) -> None:
         """
@@ -474,15 +516,13 @@ class DSLibrary(object):
         """
         use_mlflow = self._mlflow_all
         # map resource name using supplied specs
-        resource_spec = kwargs
-        resource_spec.update((self._spec.get("outputs") or {}).get(resource_name) or {})
-        resource_name_xlt = resource_spec.pop("uri", None) or resource_name
-        file_extension = os.path.splitext(resource_name_xlt)[1].strip(".")
+        uri, kwargs = self._xlt_resource(resource_name, _mode="outputs", **kwargs)
+        # FIXME use the URL extension splitter!
+        file_extension = os.path.splitext(uri)[1].strip(".")
         if not file_extension:
-            self._default_format_for_aliases(resource_name, resource_spec)
-        format = resource_spec.pop("format", None) or file_extension
-        # NOTE: specs take precedence over kwargs, making model-supplied values effectively only defaults
-        format_options = dict(resource_spec.get("format_options") or {})
+            self._default_format_for_aliases(resource_name, kwargs)
+        format = kwargs.pop("format", None) or file_extension
+        format_options = dict(kwargs.get("format_options") or {})
         # detect SQL target
         if format_options.get("sql_table"):
             # TODO formulate a bulk INSERT to write to the target table
@@ -506,14 +546,20 @@ class DSLibrary(object):
             if append:
                 try:
                     # verify file exists, capture column names
-                    with opener(resource_name_xlt, 'rb') as f_r:
+                    with opener(uri, 'rb') as f_r:
                         chk_args = {"nrows": 0}
                         if format == "json":
                             chk_args["lines"] = True
+                        # apply write args to read format
+                        for f in ["sep"]:
+                            if f in format_options:
+                                chk_args[f] = format_options[f]
                         df0 = PANDAS_READ_FORMATS[df_format](f_r, **chk_args)
                         # verify column names match
-                        if set(df0.columns) != set(content.columns):
-                            raise DSLibraryException(f"Cannot append to {resource_name}, column names do not match")
+                        cols_l = set(df0.columns)
+                        cols_r = set(content.columns)
+                        if cols_l != cols_r:
+                            raise DSLibraryException(f"Cannot append to {resource_name}, column names do not match: {', '.join(sorted(cols_l))} vs {', '.join(sorted(cols_r))}")
                 except FileNotFoundError:
                     append = False
             mode = "w"
@@ -525,7 +571,7 @@ class DSLibrary(object):
                 format_options["index"] = False
             if format == "json":
                 format_options.update({"orient": "records", "lines": True})
-            with opener(resource_name_xlt, mode) as f_w:
+            with opener(uri, mode) as f_w:
                 df_method(f_w, **format_options)
             return
         # save matplotlib figures as images
@@ -533,7 +579,7 @@ class DSLibrary(object):
             if append:
                 raise ValueError("append is not supported for figures")
             if use_mlflow:
-                mlflow.log_figure(content, resource_name_xlt)
+                mlflow.log_figure(content, uri)
                 return
             buf = io.BytesIO()
             image_format = format or "png"
@@ -553,14 +599,14 @@ class DSLibrary(object):
             if use_mlflow:
                 if not isinstance(content, str):
                     content = content.decode("utf-8")
-                mlflow.log_text(content, resource_name_xlt)
+                mlflow.log_text(content, uri)
             else:
                 with self.open_resource(resource_name, mode=mode_base + ('' if isinstance(content, str) else 'b')) as f_w:
                     f_w.write(content)
         elif isinstance(content, (dict, list, tuple)):
             # write yaml or json
             if use_mlflow and isinstance(content, dict):
-                mlflow.log_dict(content, resource_name_xlt)
+                mlflow.log_dict(content, uri)
                 return
             if format in ("yaml", "yml"):
                 with self.open_resource(resource_name, mode=mode_base) as f_w:
@@ -604,6 +650,29 @@ class DSLibrary(object):
         with self.open_resource(resource_name, mode=mode) as f_r:
             return f_r.read()
 
+    @staticmethod
+    def _in_memory_append(a, b):
+        if isinstance(a, str):
+            return a + str(b)
+        if isinstance(a, (bytes, bytearray)):
+            return a + (b if isinstance(b, (bytes, bytearray)) else str(b).encode("utf-8"))
+        if hasattr(a, "append"):
+            return a.append(b)
+        if isinstance(a, list):
+            return a + list(b)
+        if isinstance(a, dict) and isinstance(b, dict):
+            out = dict(a)
+            out.update(b)
+            return out
+        return b
+
+    def _capturing_output(self, resource_name: str) -> bool:
+        """
+        Whether the given output is supposed to be captured.  This is part of the development/debugging feature that
+        frees a caller from having to store i/o data in files to test a model.
+        """
+        return ((self._spec.get("outputs") or {}).get(resource_name) or {}).get("capture")
+
     def write_resource(self, resource_name: str, content=None, from_filename: str=None, append: bool=False, **kwargs):
         """
         Write various types of content or files to an output resource.
@@ -617,6 +686,16 @@ class DSLibrary(object):
         :param append:              True to append, False to replace.
         """
         if content is not None:
+            # content can be requested to be captured
+            if self._capturing_output(resource_name):
+                if "capture" not in self._spec:
+                    self._spec["capture"] = {}
+                if append:
+                    prev = self._spec["capture"].get(resource_name)
+                    if prev is not None:
+                        content = self._in_memory_append(prev, content)
+                self._spec["capture"][resource_name] = content
+                return
             self._write_content(self.open_resource, resource_name, content=content, append=append, **kwargs)
         elif from_filename is not None:
             use_mlflow = self._mlflow_all
@@ -642,17 +721,16 @@ class DSLibrary(object):
         :param resource_name:   The name of an input resource, or of a local file, or the URI of an external resource.
                 For sql- and nosql- type data sources, ":<table_name>" is expected.
         """
-        # map requested resource through input spec
-        resource_spec = (self._spec.get("inputs") or {}).get(resource_name) or {}
-        resource_name_xlt = resource_spec.get("uri") or resource_name
-        # mix in format arguments supplied in input specification
-        format_options = dict(resource_spec)
-        format_options.pop("uri", None)
-        kwargs.update(format_options)
+        # check for specifically supplied data
+        supplied = (self._spec.get("data") or {}).get(resource_name)
+        if supplied is not None and hasattr(supplied, "columns"):
+            return supplied
+        # map requested resource (resource_name) through input spec and override defaults (kwargs)
+        uri, open_args = self._xlt_resource(resource_name, **kwargs)
         # detect SQL
-        if kwargs.get("sql") or kwargs.get("sql_table"):
-            sql = kwargs.pop("sql", None)
-            table = kwargs.pop("sql_table", None)
+        if open_args.get("sql") or open_args.get("sql_table"):
+            sql = open_args.pop("sql", None)
+            table = open_args.pop("sql_table", None)
             if table:
                 sql = f"SELECT * from {sql_enquote_id(table)}"
             db_conn = self.get_sql_connection(resource_name)
@@ -664,19 +742,20 @@ class DSLibrary(object):
             finally:
                 db_conn.close()
             return df
-        if kwargs.get("nosql_collection"):
-            collection = kwargs["nosql_collection"]
+        if open_args.get("nosql_collection"):
+            collection = open_args["nosql_collection"]
             # TODO code get_nosql_connection() then implement accordingly
             raise DSLibraryException("Loading from nosql is not supported yet.")
         # TODO support read from time series database
         # TODO we should offer a validation option that checks the data types against the schema
         # format can be specified, or file extension will be used
-        file_extension = os.path.splitext(resource_name_xlt)[1].strip(".")
+        # FIXME use the URL splitter, not this one!
+        file_extension = os.path.splitext(uri)[1].strip(".")
         if not file_extension:
-            self._default_format_for_aliases(resource_name, kwargs)
-        format = kwargs.pop("format", None) or file_extension
+            self._default_format_for_aliases(resource_name, open_args)
         # options for read function can be given
-        format_options = kwargs.pop("format_options", None) or {}
+        format = open_args.pop("format", None) or file_extension
+        format_options = open_args.pop("format_options", None) or {}
         # various formats and their read function
         if format in PANDAS_READ_FORMATS:
             with self.open_resource(resource_name, mode='rb', **kwargs) as f_r:
@@ -701,13 +780,15 @@ class DSLibrary(object):
         """
         Un-pickle a saved model from the default model binary, or from one of its named parts.
         """
-        return pickle.load(self.open_model_binary(part=part, mode='rb'))
+        with self.open_model_binary(part=part, mode='rb') as f_r:
+            return pickle.load(f_r)
 
     def save_pickled_model(self, model_object, part: str = None):
         """
         Use pickle to save a model.
         """
-        return pickle.dump(model_object, self.open_model_binary(part=part, mode='wb'))
+        with self.open_model_binary(part=part, mode='wb') as f_w:
+            return pickle.dump(model_object, f_w)
 
     def read_run_data(self, resource_name: str, mode='rb') -> (str, bytes):
         """
