@@ -75,6 +75,28 @@ class DSLibraryViaMMLibrary(DSLibrary):
                 raise FileNotFoundError(f"not found: {path}")
         raise DSLibraryException(f"Unsupported mode: {mode}")
 
+    def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
+        """
+        The model-binary is just one file, but we can support 'part' by assuming it is a ZIP file.
+        """
+        if part:
+            import zipfile
+            if mode in ("w", "wb", "a", "ab"):
+                # we can't write parts because each call to mmlibrary.new_version() creates a new trained model
+                raise ValueError(f"Unsupported mode for model binary with parts: {mode}")
+            with self.open_model_binary() as f_r:
+                with zipfile.ZipFile(f_r) as zf:
+                    if mode == "r":
+                        return io.StringIO(zf.read(part).decode("utf-8"))
+                    if mode == "rb":
+                        return io.BytesIO(zf.read(part))
+        if "r" in mode:
+            get_model = self._find_method("get_model", "getModel")
+            return open(get_model(), mode)
+        if "w" in mode:
+            wr_model = self._find_method("new_version", "newVersion")
+            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=lambda f_r: wr_model(f_r.read()))
+
     def log_metric(self, metric_name: str, metric_value: float, step: int=0) -> None:
         """
         MMlibrary has its own KPI logging method.
